use api::{Item, Transaction};
use dioxus::prelude::*;
use dioxus_material_icons::{MaterialIcon, MaterialIconStylesheet};
use dioxus_sdk::storage::use_persistent;
use qrscan::QRCodeScanPage;
use setup::{Credentials, SetupPage};

pub mod api;
pub mod page;
pub mod qrscan;
pub mod setup;

use page::{supply::SupplyPageParam, *};

#[derive(Debug, Clone, Routable, PartialEq)]
#[rustfmt::skip]
enum Route {
    #[layout(Navbar)]
    #[route("/")]
    Home {},
    #[route("/items")]
    ItemPage {},
    #[route("/flows")]
    FlowPage {},
    #[route("/locations")]
    LocationsPage {},

    #[route("/item/:id")]
    ItemDetailPage { id: String },
    #[route("/transaction/:id")]
    TransactionPage { id: String },

    #[route("/item/:item/:variant/consume/:id")]
    ConsumePage { id: String, item: String, variant: String },


    #[route("/item/:item/supply?:param")]
    SupplyPage { item: String, param: SupplyPageParam }
//    #[route("/blog/:id")]
//    Blog { id: i32 },

    // Pages
    // 
}

const FAVICON: Asset = asset!("/assets/favicon.ico");
const MAIN_CSS: Asset = asset!("/assets/main.css");
const TAILWIND_CSS: Asset = asset!("/assets/tailwind.css");

pub static API: GlobalSignal<Option<api::API>> = Signal::global(|| None);

fn main() {
    dioxus_sdk::set_dir!();
    dioxus::launch(App);
}

#[component]
fn App() -> Element {
    let creds = use_persistent("creds", || Credentials::default());

    spawn(async move {
        let res = use_persistent("creds", || Credentials::default());
        if !res.read().empty() {
            let api = api::API::new(res.read().instance_url.clone()).await;
            *crate::API.write() = Some(api);
        }
    });

    rsx! {
        document::Link { rel: "icon", href: FAVICON }
        document::Link { rel: "stylesheet", href: MAIN_CSS }
        document::Link { rel: "stylesheet", href: TAILWIND_CSS }
        MaterialIconStylesheet { }

        if creds.read().empty() {
            SetupPage {  }
        } else {
            Router::<Route> {}
        }
    }
}

#[component]
fn ItemTile(item: Item) -> Element {
    rsx! {
        Link {
                to: Route::ItemDetailPage { id: item.uuid },
                {item.name.as_str()}
            }
    }
}

#[component]
fn FlowPage() -> Element {
    rsx! {}
}

/// Home page
#[component]
fn QRTest() -> Element {
    let found = use_signal(|| String::new());
    let mut showQR = use_signal(|| false);

    rsx! {

        if !found.read().is_empty() {
            p {
                {format!("FOUND {}", found.read())}
            }
        },

        input {
            type: "text"
        }

        if *showQR.read() {
            QRCodeScanPage { result: found, show: showQR }
        },

                button {
            onclick: move |_| {
                showQR.set(true);
            },
            "Click"
        }

    }
}

#[component]
fn TransactionCard(t: Transaction) -> Element {
    rsx! {
    // TODO : transaction card
    Link {
        to: Route::TransactionPage { id: t.uuid.clone() },
        p { {t.uuid.as_str()} }
    }
        }
}

/// Shared navbar component.
#[component]
fn Navbar() -> Element {
    rsx! {
        div {
            id: "navbar",
            Link {
                to: Route::Home {},
                div {
                    class: "flex flex-row gap-2 p-2",
                    MaterialIcon { name: "home", size: 24 },
                    "Home"
                }
            }
            Link {
                to: Route::ItemPage {  },
                div {
                    class: "flex flex-row gap-2 p-2",
                    MaterialIcon { name: "category", size: 24 },
                    "Items"
                }
            }
            Link {
                to: Route::FlowPage {  },
                div {
                    class: "flex flex-row gap-2 p-2",
                    MaterialIcon { name: "assignment", size: 24 },
                    "Flows"
                }
            }
            Link {
                to: Route::LocationsPage {  },
                div {
                    class: "flex flex-row gap-2 p-2",
                    MaterialIcon { name: "map", size: 24 },
                    "Locations"
                }
            }
        }

        Outlet::<Route> {}
    }
}
