use dioxus::prelude::*;
use dioxus_sdk::storage::use_persistent;
use serde::{Deserialize, Serialize};
use std::collections::HashMap;

#[derive(Default, Serialize, Deserialize, Clone, PartialEq)]
pub struct Credentials {
    pub instance_url: String,
    pub token: String,
}

impl Credentials {
    pub fn empty(&self) -> bool {
        self.instance_url.is_empty() && self.token.is_empty()
    }
}

#[derive(Default, Clone)]
struct FormErrors {
    instance_url: Option<String>,
    token: Option<String>,
}

#[component]
pub fn SetupPage() -> Element {
    let mut form_data = use_signal(Credentials::default);
    let mut form_data2 = form_data.clone();
    let mut form_errors = use_signal(FormErrors::default);
    let submitted = use_signal(|| false);

    let mut validate_form = move || {
        let mut errors = FormErrors::default();
        if form_data2.read().instance_url.is_empty() {
            errors.instance_url = Some("Instance URL is required".to_string());
        }
        if form_data2.read().token.is_empty() {
            errors.token = Some("Token is required".to_string());
        }

        form_errors.set(errors.clone());
        errors.instance_url.is_none() && errors.token.is_none()
    };

    let on_submit = move |evt: FormEvent| {
        evt.prevent_default();

        if validate_form() {
            let mut creds = use_persistent("creds", || Credentials::default());

            spawn(async move {
                let api = crate::api::API::new(creds.read().instance_url.clone()).await;
                *crate::API.write() = Some(api);
            });

            creds.set(form_data.read().clone());
        }
    };

    rsx! {
        div {
            h1 { "Setup Page" }
            form {
                onsubmit: on_submit,
                div {
                    label { "Instance URL:" }
                    input {
                        value: "{form_data.read().instance_url}",
                        oninput: move |evt| {
                            let mut form_data = form_data.write();
                            form_data.instance_url = evt.value();
                        }
                    },
                    {form_errors.read().instance_url.as_ref().map(|error| rsx! {
                        p { class: "error", "{error}" }
                    })}
                }
                div {
                    label { "Token:" }
                    input {
                        r#type: "password",
                        value: "{form_data.read().token}",
                        oninput: move |evt| {
                            let mut form_data = form_data.write();
                            form_data.token = evt.value();
                        },
                    }
                    {form_errors.read().token.as_ref().map(|error| rsx! {
                        p { class: "error", "{error}" }
                    })}
                }
                button { type: "submit", "Complete Setup" }
            }
            if *submitted.read() {
                div {
                    h3 { "Setup Complete!" }
                }
            }
        }
    }
}
